% spiky signal in which half the spikes are O(1) and the remainder are
% 80-100 dB below. The signal also contains noise at 120 dB below.

clear all

global n;
global k;
global dct_rows;

addpath('../gpu_warmup');

% Solve it on GPU? (0=no)
solve_GPU = 1;
% Solve it on Host? (0=no)
solve_host = 1;

%  spikes with DCT sensing matrix

% initialize random number generators
rand('seed',9123191); randn('seed',67812687681);


% n is the original signal length (factor of 2)
%n = 2^12;
n = 2^16;
%n = 2^20;

% k is number of observations to make
k = n/8;

% pick out k random rows of the dct matrix to use
q = randperm(n); 
dct_rows = q(1:k);
dct_rows=sort(dct_rows);
dct_rows=int32(dct_rows);

% noise levels (dB below main signals)
noisefloor=120;
dynrangelo=100;
dynrangehi=80;

% construct the signal spikes
n_spikes = floor(.2*k); 
n_bigspikes = floor(n_spikes/2); 
n_smallspikes = n_spikes-n_bigspikes;

f = zeros(n,1);
q = randperm(n);
f(q(1:n_bigspikes)) = 1 + 0.1*rand(n_bigspikes,1);
scales = dynrangehi + rand(n_smallspikes,1)*(dynrangelo-dynrangehi);
scales = -log(10)*scales/20;
f(q(n_bigspikes+1:n_bigspikes+n_smallspikes)) = exp(scales);

% add some noise
f = f + 10^(-noisefloor/20)*randn(n,1);

% and construct the measurements
y = dctA(f);

% noisy observations?
%sigma = 0.0001;
sigma=0.0;
y = y + sigma*randn(k,1);

% regularization parameter
tau_max = max(abs(dctAT(y)));
tau_vals = [0.0001 0.000033 0.00001]*tau_max;

fprintf(1,'\n\n-------------------------------------------------\n')   
fprintf(1,'-------------------------------------------------\n')   
fprintf(1,'Problem: n = %g,  k = %g, number of spikes = %g\n',n,k,n_spikes)
fprintf(1,'-------------------------------------------------\n')

% set convergence tolerance
stop_tol = 1.e-6;

% choose the number of continuation steps (1=no continuation)
continuationSteps=10;

n_runs = length(tau_vals);
iter_gpu  = zeros(n_runs,1);
iter_host = zeros(n_runs,1);
time_gpu  = zeros(n_runs,1);
time_host = zeros(n_runs,1);
mse_gpu  = zeros(n_runs,1);
mse_host = zeros(n_runs,1);

% solve it on GPU
if solve_GPU
  fprintf('\n\n Solving on GPU:\n');
  % warmup gpu
  gpu_warmup(2);

  i=1;
  for tau = tau_vals,
    fprintf(1,'Parameters: sigma = %g, tau/tau_max = %g, debiasing = %g\n',sigma,tau/tau_max,0)
    x_start = zeros(n,1);
    time_gpu_turnaround = cputime();
    [x,iter_gpu(i),time_gpu(i),exectime_gpu] = ...
      SpaRSA_gpu(single(y), single(x_start), dct_rows,...
      single(tau), continuationSteps, single(stop_tol));
    time_gpu_turnaround = cputime() - time_gpu_turnaround;
    % calculate final mse
    mse_gpu(i) = (1/n)*norm(f-x)^2;
    fprintf(1,' final MSE = %6.3e, iterations = %d\n', mse_gpu(i), iter_gpu(i));
    fprintf(1,' turnaround time = %7.3f,  total time = %7.3f, exec time = %7.3f\n', time_gpu_turnaround, time_gpu(i), exectime_gpu);
    i = i+1;
  end
end

% solve it on host
if solve_host
  fprintf('\n\n Solving on Host:\n');
  
  i=1;
  for tau = tau_vals,
    fprintf(1,'Parameters: sigma = %g, tau/tau_max = %g, debiasing = %g\n',sigma,tau/tau_max,0)
    x_start = zeros(n,1);
    time_host(i) = cputime();
    [x,iter_host(i)] = SpaRSA_host(y, x_start, @dctA, @dctAT,...
        tau, continuationSteps, stop_tol);
    time_host(i) = cputime() - time_host(i);
    % calculate final mse
    mse_host(i) = (1/n)*norm(f-x)^2;
    fprintf(1,' final MSE = %6.3e, iterations = %d, total time = %7.3f\n', ...
        mse_host(i), iter_host(i), time_host(i));
    i = i+1;
  end
end

if solve_host && solve_GPU
  fprintf('\n\nLatex table....\n\n');
  fprintf('\\begin{table}\\centering\n');
  fprintf('\\begin{tabular}{|c|ccc|ccc|cc|}\\hline\n');
  fprintf('\\multirow{2}{*}{$\\tau/\\taumax$} & \\multicolumn{3}{|c|}{CPU} & \\multicolumn{3}{|c|}{GPU} & \\multicolumn{2}{|c|}{Speedup}\\\\\n');
  fprintf('\\cline{2-9}\n')
  fprintf(' & iters & time (s) & MSE & iters & time (s) & MSE & total & per iter\\\\\n');
  fprintf('\\hline\n');
  i = 1;
  for tau = tau_vals,
    fprintf('%f & %d & %.2f & %.1e & %d & %.2f & %.1e & %.0f & %.0f\\\\\n', tau/tau_max, iter_host(i), time_host(i), mse_host(i), iter_gpu(i), time_gpu(i), mse_gpu(i), time_host(i)/time_gpu(i), (time_host(i)*iter_gpu(i))/(time_gpu(i)*iter_host(i))); 
    i = i+1;
  end
  fprintf('\\hline\n');
  fprintf('\\end{tabular}\n');
  fprintf('\\caption{Computational results for a 1-D DCT sensing matrix of dimension $%d\\times %d$, with %d spikes\\label{ta:1D.%d}}\n', k, n, n_spikes, n)
  fprintf('\\end{table}\n\n');
%    fprintf('\nSpeedup          %.2f\n', time_host/exectime_gpu);
%    fprintf('Speedup per iter %.2f\n', (time_host*iter_gpu)/(exectime_gpu*iter));
end

exit
