%  spikes in 2D, sensing matrix is a composition of 2d-DCT with random
%  observation
close all
clear all
%clf

global dct_rows
global n
global m

addpath('../gpu_warmup');

% Solve it on GPU? (0=no)
solve_GPU = 1;
% Solve it on Host? (0=no)
solve_host = 1;
% Plot results?
plotresults = 0;

% signal length
logm = 8;
%logm = 10;
%logm = 11;
m = 2^logm;
n=m*m;
% proportion of spikes
p_spikes = 0.001;
% number of observations to make
k = round(20*p_spikes*n);

q= randperm(n);
dct_rows = q(1:k);
dct_rows = sort(dct_rows);
dct_rows = int32(dct_rows);

% set seeds to solve the same problem each time
rand('seed',91823811); randn('seed',67216877);

% random +/- 1 signal
f = sign(rand(m,m)-0.5).*(rand(m,m) <= p_spikes);
f_exact = f;
nSpikes = length(find(f(:)~=0));

if plotresults
  scrsz = get(0,'ScreenSize');
  set(1,'Position',[10 scrsz(4)*0.05 scrsz(3)/4 0.85*scrsz(4)])
  figure(1)
  subplot(2,1,1)
  imagesc(f)
  colormap(gray)
  axis off
  axis equal
  title('Original image','FontName','Times','FontSize',14)
end


% create observation operator; in this case
% it will be a blur function
disp('Creating observation operator...');

% generate noisy observations
sigma = 0.001;
y = dctA(f) + sigma*randn(k,1);

% regularization parameter
Aty = dctAT(y);
tau_max = norm(Aty(:),inf);
tau_vals = [.1 .05 .02]*tau_max;

stop_tol = 1.e-6;
continuationSteps=10;

n_runs = length(tau_vals);
iter_gpu  = zeros(n_runs,1);
iter_host = zeros(n_runs,1);
time_gpu  = zeros(n_runs,1);
time_host = zeros(n_runs,1);
mse_gpu  = zeros(n_runs,1);
mse_host = zeros(n_runs,1);

% solve it on GPU
if solve_GPU
  fprintf('\n\nSolve on GPU with %d continuation steps\n', continuationSteps);
  % warmup gpu
  gpu_warmup(2);

  i=1;
  for tau=tau_vals,
    fprintf('%d x %d pixels, %d spikes, %d observations. Noise=%6.2e, tau/tau_max=%6.2e\n', ...
             m, m, nSpikes, k, sigma, tau/norm(Aty(:),inf));
    f_start = zeros(size(f));
    time_gpu_turnaround = cputime();
    % solve it
    [f, iter_gpu(i), time_gpu(i), exectime_gpu] = SpaRSA_gpu(single(y),single(f_start),dct_rows,single(tau),...
                                                             continuationSteps,single(stop_tol));
    time_gpu_turnaround = cputime()-time_gpu_turnaround;
    mse_gpu(i) = sum((f(:)-f_exact(:)).^2) / size(f(:),1);
    fprintf(' mean square error: %8.2e, iterations = %d\n', mse_gpu(i), iter_gpu(i));
    fprintf(' turnaround time = %7.3f, total time = %7.3f, exec time = %7.3f\n\n', time_gpu_turnaround, time_gpu(i), exectime_gpu);
    i=i+1;
  end
end

% solve it on host
if solve_host
  fprintf('\n\nSolve on host with %d continuation steps\n', continuationSteps);

  i=1;
  for tau=tau_vals,
    fprintf('%d x %d pixels, %d spikes, %d observations. Noise=%6.2e, tau/tau_max=%6.2e\n', ...
             m, m, nSpikes, k, sigma, tau/norm(Aty(:),inf));
    f_start = zeros(size(f));
    time_host(i) = cputime();
    [f,iter_host(i)] = SpaRSA_host(y, f_start, @dctA, @dctAT,...
                                 tau, continuationSteps, stop_tol);
    time_host(i) = cputime()-time_host(i);
    mse_host(i) = sum((f(:)-f_exact(:)).^2) / size(f(:),1);
    fprintf(' mean square error: %8.2e, iterations = %d\n', mse_host(i), iter_host(i));
    fprintf(' total time = %7.3f\n\n', time_host(i));
    i=i+1;
  end
end

if solve_host && solve_GPU
  fprintf('\n\nLatex table....\n\n');
  fprintf('\\begin{table}\\centering\n');
  fprintf('\\begin{tabular}{|c|ccc|ccc|cc|}\\hline\n');
  fprintf('\\multirow{2}{*}{$\\tau/\\taumax$} & \\multicolumn{3}{|c|}{CPU} & \\multicolumn{3}{|c|}{GPU} & \\multicolumn{2}{|c|}{Speedup}\\\\\n');
  fprintf('\\cline{2-9}\n')
  fprintf(' & iters & time (s) & MSE & iters & time (s) & MSE & total & per iter\\\\\n');
  fprintf('\\hline\n');
  i = 1;
  for tau = tau_vals,
    fprintf('%f & %d & %.2f & %.1e & %d & %.2f & %.1e & %.0f & %.0f\\\\\n', tau/tau_max, iter_host(i), time_host(i), mse_host(i), iter_gpu(i), time_gpu(i), mse_gpu(i), time_host(i)/time_gpu(i), (time_host(i)*iter_gpu(i))/(time_gpu(i)*iter_host(i))); 
    i = i+1;
  end
  fprintf('\\hline\n');
  fprintf('\\end{tabular}\n');
  fprintf('\\caption{Computational results for a 2-D DCT sensing matrix of dimension $%d\\times %d$, with %d spikes\\label{ta:2D.%d}}\n', k, n, nSpikes, n)
  fprintf('\\end{table}\n\n');
%    fprintf('\nSpeedup          %.2f\n', time_host/exectime_gpu);
%    fprintf('Speedup per iter %.2f\n', (time_host*iter_gpu)/(exectime_gpu*iter));
end

if plotresults
  % ================= Plotting results ==========
  figure(1)
  subplot(2,1,2)
  imagesc(f)
  colormap(gray)
  axis off
  axis equal
  title('Reconstructed image','FontName','Times','FontSize',14);
end

