/*
 * function [u, w1,w2, TimeCost, itr] = TV_PDHDB(w1,w2,f,h,lbd,NIT,GapTol);
 *
 * GPU (CUDA) implementation of image deblurring algorithm, PDHG-B.
 *
 * Author: Sangkyun Lee
 * Last Modified: 2008/11/7
 */

#include <time.h>
#include "mex.h"
#include "cudaMath.cu"

#define MATLAB
//#define SAFE_MODE
//#define VERBOSE

void print_usage()
{
    debug("[u,w1,w2,TimeCost,itr] = TV_PDHDB(w1,w2,f,h,lbd,NIT,GapTol)\n");
}

/*
 * For two real variables u and DivW, compute (u-DivW) and convert the result to a complex variable
 */
__global__ void __minusMatMatToComplex(Complex* cpx_u, const float* u, const float* DivW, float theta, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;
    Complex c;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	c.x = u[lin_idx] - theta*DivW[lin_idx];
        c.y = 0.f;
        cpx_u[lin_idx] = c;
    }
}

/*
 * Compute the real part of (fft./denom)
 */
__global__ void __normalizedRealPart (float* out, const Complex* fft, float denom, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
        out[lin_idx] = __fdividef(fft[lin_idx].x, denom);
    }
}

/*
 * Compute hh = conj(fft_h).*fft_h,
 * and fft_chf = conj(fft_h).*fft_f
 */
//__conj<<<dimGrid, dimBlock>>> (fft_chf, fft_h, m, n);
//__timesMatMatC<<<dimGrid, dimBlock>>> (hh, fft_chf, fft_h, m, n);		// hh = conj(h).*h
//__timesMatMatC<<<dimGrid, dimBlock>>> (fft_chf, fft_chf, fft_f, m, n);	// chf = conj(h).*f;
__global__ void __computeCHFandHH(Complex* fft_chf, Complex* hh, const Complex* fft_f, const Complex* fft_h, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;
    Complex c_hh, c_chf, c1, c2;

    if(X < m && Y < n) {
	lin_idx = __imul(m,Y) + X;
	c1 = fft_h[lin_idx];
	c2 = fft_f[lin_idx];

	// compute hh
	c_hh.x = __fmul_rn(c1.x,c1.x) - __fmul_rn(-c1.y,c1.y);
	c_hh.y = __fmul_rn(-c1.y,c1.x) + __fmul_rn(c1.x,c1.y);

	// compute chf
	c_chf.x = __fmul_rn(c1.x,c2.x) - __fmul_rn(-c1.y,c2.y);
	c_chf.y = __fmul_rn(-c1.y,c2.x) + __fmul_rn(c1.x,c2.y);

	fft_chf[lin_idx] = c_chf;
	hh[lin_idx] = c_hh;
    }
}

/*
 * Projection
 */
__global__ void __normalize(FLOAT* w1, FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx = m*Y + X;

    FLOAT w1_e = w1[lin_idx];
    FLOAT w2_e = w2[lin_idx];
    FLOAT norm = hypotf(w1_e, w2_e);

    if(norm > 1.0f)
    {
	w1[lin_idx] = __fdividef(w1_e, norm);
	w2[lin_idx] = __fdividef(w2_e, norm);
    }
}

/*
 * Update dual variable u
 */
/// new_u = (fft2(u - theta*DivW) + theta*lbd*(ch.*fft_f) ) ./ (1 + theta*lbd*hh);
__global__  void __updateDual(Complex* u, const Complex* u_upd, const Complex* fft_chf, const Complex* hh, float w, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;
    float denom;
    Complex c, c_upd, c_chf, c_hh;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;

	c_upd = u_upd[lin_idx];
	c_chf = fft_chf[lin_idx];
	c_hh = hh[lin_idx];

	c_hh.x = 1.f + w*c_hh.x;
	c_hh.y = w*c_hh.y;
	c_upd.x += w*c_chf.x;
	c_upd.y += w*c_chf.y;
	
	denom = __fsqr_rn(c_hh.x) + __fsqr_rn(c_hh.y);
	//denom = c_hh.x*c_hh.x + c_hh.y*c_hh.y;
	c.x = __fdividef(c_upd.x*c_hh.x + c_upd.y*c_hh.y, denom);
	c.y = __fdividef(c_upd.y*c_hh.x - c_upd.x*c_hh.y, denom);
	u[lin_idx] = c;
    }
}


/*
 * Main Entry Function for Matlab
 *
 */
//function [u, w1,w2, TimeCost, itr] = TV_PDHDB(w1,w2,f,h,lbd,NIT,GapTol);
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
    FLOAT *w1_h, *w2_h, *f_h, *h_h;			// in host
    FLOAT *w1, *w2, *f, *h, *DivW;	// in device
    Complex *cpx_u, *fft_tmp, *fft_f, *fft_h, *fft_chf, *hh;
    FLOAT *u, *ux, *uy;
    FLOAT *tmp;				// for pointer swapping
    FLOAT lbd, GapTol, tol;
    int NIT;
    int m,n,mn;
    int memsize, cpx_memsize;

    // For FFT on GPU
    cufftHandle fft_plan;

    // output variables
    FLOAT *u_out, *w1_out, *w2_out;

    // opt. loop variables
    double total_time;
    int itr;
    FLOAT tau, theta;

    total_time = cputime();

    if(nrhs != 7) {
	print_usage();
	return;
    }

    // input variables
    w1_h = (FLOAT*) mxGetData(prhs[0]);		// these arguments should be single()
    w2_h = (FLOAT*) mxGetData(prhs[1]);
    f_h = (FLOAT*) mxGetData(prhs[2]);
    h_h = (FLOAT*) mxGetData(prhs[3]);		// shifted padded real-valued kernel
    lbd = (FLOAT) mxGetScalar(prhs[4]);
    NIT = (int) mxGetScalar(prhs[5]);
    GapTol = (FLOAT) mxGetScalar(prhs[6]);

    m = mxGetM(prhs[2]);
    n = mxGetN(prhs[2]);
    mn = m*n;
    memsize = sizeof(FLOAT)*mn;
    cpx_memsize = sizeof(Complex)*mn;

    // set device
    CUDA_CALL( cudaSetDevice(2) );	// 9800
    cudaMath_init();

    // allocate GPU memory
    CUDA_CALL( cudaMalloc((void**)&w1, memsize) );
    CUDA_CALL( cudaMalloc((void**)&w2, memsize) );
    CUDA_CALL( cudaMalloc((void**)&f, memsize) );
    CUDA_CALL( cudaMalloc((void**)&h, memsize) );
    CUDA_CALL( cudaMalloc((void**)&DivW, memsize) );
    CUDA_CALL( cudaMalloc((void**)&u, memsize) );
    CUDA_CALL( cudaMalloc((void**)&ux, memsize) );
    CUDA_CALL( cudaMalloc((void**)&uy, memsize) );

    CUDA_CALL( cudaMalloc((void**)&cpx_u, cpx_memsize) );
    CUDA_CALL( cudaMalloc((void**)&fft_f, cpx_memsize) );
    CUDA_CALL( cudaMalloc((void**)&fft_h, cpx_memsize) );
    CUDA_CALL( cudaMalloc((void**)&fft_chf, cpx_memsize) );
    CUDA_CALL( cudaMalloc((void**)&fft_tmp, cpx_memsize) );
    CUDA_CALL( cudaMalloc((void**)&hh, cpx_memsize) );

    // copy input vectors to GPU memory
    CUDA_CALL( cudaMemcpy(f, f_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(h, h_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(w1, w1_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(w2, w2_h, memsize, cudaMemcpyHostToDevice) );
//    CUDA_CALL( cudaMemset(w1, 0, memsize) );
//    CUDA_CALL( cudaMemset(w2, 0, memsize) );
    CUDA_CALL( cudaMemcpy(u, f, memsize, cudaMemcpyDeviceToDevice) );

    // bind textures
    CUDA_CALL( cudaBindTexture(0, tex, u, memsize) );
    CUDA_CALL( cudaBindTexture(0, tex1, w1, memsize) );
    CUDA_CALL( cudaBindTexture(0, tex2, w2, memsize) );

    // define block and grid
    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(idivup(m,BLK_LEN_2D), idivup(n,BLK_LEN_2D));

    // Generate fft_f, fft_h and fft_chf
    CUFFT_CALL( cufftPlan2d( &fft_plan, m, n, CUFFT_C2C ) );

    __toComplex<<<dimGrid, dimBlock>>> (fft_tmp, f, m, n);
    // CUDA_CHECK_ERROR("__toComplex error");
    CUFFT_CALL( cufftExecC2C( fft_plan, (cufftComplex*)fft_tmp, (cufftComplex*)fft_f, CUFFT_FORWARD) );
    __toComplex<<<dimGrid, dimBlock>>> (fft_tmp, h, m, n);
    CUFFT_CALL( cufftExecC2C( fft_plan, (cufftComplex*)fft_tmp, (cufftComplex*)fft_h, CUFFT_FORWARD) );
    //__conj<<<dimGrid, dimBlock>>> (fft_chf, fft_h, m, n);
    //__timesMatMatC<<<dimGrid, dimBlock>>> (hh, fft_chf, fft_h, m, n);		// hh = conj(h).*h
    //__timesMatMatC<<<dimGrid, dimBlock>>> (fft_chf, fft_chf, fft_f, m, n);	// chf = conj(h).*f;
    __computeCHFandHH<<<dimGrid, dimBlock>>> (fft_chf, hh, fft_f, fft_h, m, n);

    //
    // Algorithm begins.
    //
    gradient2(ux, uy, u, m, n);

for (itr=1; itr<=NIT; ++itr)
{
    // Solve dual
    // choose tau
    tau = 10. + 40.*itr;
    __update<<<dimGrid, dimBlock>>> (w1, ux, tau, m, n);	// w1 = w1 + tau*ux;
    __update<<<dimGrid, dimBlock>>> (w2, uy, tau, m, n);	// w2 = w2 + tau*uy;
    __normalize<<<dimGrid, dimBlock>>> (w1, w2, m, n);
    divergence2(DivW, w1, w2, m, n);

    // Solve primal
    theta = (1.-.2/itr)/tau;
    
    /* Perform:
     *     new_u = real( ifft2( (fft2(u - theta*DivW) + theta*lbd*(ch.*fft_f) ) ./ (1 + theta*lbd*hh) ) );
     * is equal to call:
     *     __minusMatMat<<<dimGrid, dimBlock>>> (new_u, u, 1.f, DivW, theta, m, n);
     *     __toComplex<<<dimGrid, dimBlock>>> (cpx_u, new_u, m, n);
     * Instead, call a single function to do both for efficiency:
     */
    __minusMatMatToComplex<<<dimGrid, dimBlock>>> (cpx_u, u, DivW, theta, m, n);

    // forward FFT
    CUFFT_CALL( cufftExecC2C( fft_plan, (cufftComplex*) cpx_u, (cufftComplex*) fft_tmp, CUFFT_FORWARD) );

    __updateDual<<<dimGrid, dimBlock>>> (cpx_u, fft_tmp, fft_chf, hh, theta*lbd, m, n);

    // inverse FFT
    CUFFT_CALL( cufftExecC2C( fft_plan, (cufftComplex*) cpx_u, (cufftComplex*) fft_tmp, CUFFT_INVERSE) );

    //__rdivideMatConstC<<<dimGrid,dimBlock>>> (fft_tmp, fft_tmp, m*n, m, n);
    //__realPart<<<dimGrid, dimBlock>>> (new_u, fft_tmp, m, n);
    __normalizedRealPart<<<dimGrid, dimBlock>>> (ux, fft_tmp, mn, m, n);	// ux is u_new

    __minusMatMat<<<dimGrid, dimBlock>>> (u, u, ux, m, n);
    tol = norm_inf(u, mn) / 256.f;

    //CUDA_CALL( cudaMemcpy(u, new_u, memsize, cudaMemcpyDeviceToDevice) );
    tmp = u; u = ux; ux = tmp;

#ifdef VERBOSE
    debug(" itr %4d: tol=%5.2e\n", itr, tol);
#endif

    if ( tol < GapTol )
	break;

    gradient2(ux, uy, u, m, n);
}

    // handle return values to Matlab
    plhs[0] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    plhs[1] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    plhs[2] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    u_out = (FLOAT*) mxGetData(plhs[0]);
    w1_out = (FLOAT*) mxGetData(plhs[1]);
    w2_out = (FLOAT*) mxGetData(plhs[2]);

    CUDA_CALL( cudaMemcpy(u_out, u, memsize, cudaMemcpyDeviceToHost) );
    CUDA_CALL( cudaMemcpy(w1_out, w1, memsize, cudaMemcpyDeviceToHost) );
    CUDA_CALL( cudaMemcpy(w2_out, w2, memsize, cudaMemcpyDeviceToHost) );

    plhs[3] = mxCreateDoubleScalar((double) itr);

    // free GPU memory
    CUDA_CALL( cudaFree(w1) );
    CUDA_CALL( cudaFree(w2) );
    CUDA_CALL( cudaFree(f) );
    CUDA_CALL( cudaFree(h) );
    CUDA_CALL( cudaFree(DivW) );
    CUDA_CALL( cudaFree(u) );
    CUDA_CALL( cudaFree(ux) );
    CUDA_CALL( cudaFree(uy) );

    CUDA_CALL( cudaFree(cpx_u) );
    CUDA_CALL( cudaFree(fft_f) );
    CUDA_CALL( cudaFree(fft_h) );
    CUDA_CALL( cudaFree(fft_chf) );
    CUDA_CALL( cudaFree(fft_tmp) );
    CUDA_CALL( cudaFree(hh) );

    cudaMath_destroy();
    CUFFT_CALL( cufftDestroy( fft_plan ) );

/*
cudaStreamDestroy(stream[0]);
cudaStreamDestroy(stream[1]);
*/

    total_time = cputime() - total_time;
    plhs[4] = mxCreateDoubleScalar(total_time);

    return;
}


