clear all

addpath('../gpu_warmup');
addpath('../images');

%
% clean image
%
%I   =  imread('shape.jpg');       % 128x128
%I   =  imread('Cameraman.bmp');   % 256x256
I   =  imread('Barbara.png');     % 512x512
%I   =  imread('man.tiff');        % 1024x1024
%I   =  imread('earth2k.jpg');     % 2048x2048

%
% parameters
%
solve_GPU = 1;
solve_host = 1;
plotresults = 0;
blur_motion = 1;
blur_Gaussian = ~blur_motion;
mild_setting = 1;
severe_setting = ~mild_setting;

nit = 2000;    %max number of iterations
noise_var = 0.001^2;
lbd = min(.2/noise_var, 2e11);
%gap_tol_vals = [1.e-2 1.e-4 1.e-6];
gap_tol_vals = [1.e-3];

n_runs = length(gap_tol_vals);

time_host = zeros(n_runs,1);
time_gpu  = zeros(n_runs,1);
iter_host = zeros(n_runs,1);
iter_gpu  = zeros(n_runs,1);
mse_host  = zeros(n_runs,1);
mse_gpu   = zeros(n_runs,1);

n = length(I); 
f = double(I);

%
% Create a blur kernel
%

% motion blur kernel
if blur_motion
	if mild_setting,	h_len = 21; end
	if severe_setting,	h_len = 91; end
	h = fspecial('motion', h_len, 135);
end

% Gaussian blur kernel
if blur_Gaussian
	if mild_setting,	h_size = 21;	h_sigma = 5; end
	if severe_setting,	h_size = 41;	h_sigma = 10; end
	h = fspecial('gaussian', h_size, h_sigma);
end

% Pad and shift kernel
pad_size = size(f) - size(h);
h = padarray(h, floor(pad_size./2), 'pre');
h = padarray(h, ceil(pad_size./2), 'post');
h = fftshift(h);
h = h/sum(h(:));

% Blur image and then add noise
f = real(ifft2(fft2(h).*fft2(f)));
f = imnoise(uint8(f),'gaussian',0,noise_var); 
f = double(f);

fprintf('\nDeblurring: image size %dx%d with gaussian noise mean %.2f std %.4f\n', n, n, 0, sqrt(noise_var));

if solve_GPU
    fprintf('\nSolving on GPU,  lbd=%g\n', lbd);
    % warming-up GPU
    gpu_warmup(2);
    i=1;
    for gap_tol = gap_tol_vals,
	[u_gpu, w1,w2, iter_gpu(i), total_time] = ...
	    TV_PDHGB_gpu(single(zeros(n)),single(zeros(n)),single(f),single(h),lbd,nit,gap_tol);
	time_gpu(i) = total_time;
	singleI = single(I);
	mse_gpu(i) = sum(((u_gpu(:)-singleI(:)) ./ 256).^2) / length(I(:));
	fprintf(' gap_tol=%8g: iter=%4d, time=%8.4f, mse=%8.2e\n', gap_tol, iter_gpu(i), time_gpu(i), mse_gpu(i));
	i = i+1;
    end
end

if solve_host
    fprintf('\nSolving on host, lbd=%g\n', lbd);
    i=1;
    for gap_tol = gap_tol_vals,
	[u_host, w1,w2, time_host(i), iter_host(i)] = ...
	    TV_PDHGB_host(zeros(n),zeros(n),f,h,lbd,nit,gap_tol);
	doubleI = double(I);
	mse_host(i) = sum(((u_host(:)-doubleI(:))./ 256).^2) / length(I(:));
	fprintf(' gal_tol=%8g: iter=%4d, time=%8.4f, mse=%8.2e\n', gap_tol, iter_host(i), time_host(i), mse_host(i));
	i = i+1;
    end
end

if solve_host && solve_GPU
  fprintf('\n\nLatex table....\n\n');
  fprintf('\\begin{table}\\centering\n');
  fprintf('\\begin{tabular}{|c||cc|cc|cc|}\\hline\n');
  fprintf('\\multirow{2}{*}{Gap tol} & \\multicolumn{2}{|c|}{CPU} & \\multicolumn{2}{|c|}{GPU} & \\multicolumn{2}{|c|}{Speedup}\\\\\n');
  fprintf('\\cline{2-7}\n')
  fprintf(' & iters & time (s) & iters & time (s) & total & iter\\\\\n');
  fprintf('\\hline\n');
  i = 1;
  for gap_tol = gap_tol_vals,
    fprintf('%f & %d & %.2f & %d & %.2f & %.0f & %.0f\\\\\n', gap_tol, iter_host(i), time_host(i), iter_gpu(i), time_gpu(i), time_host(i)/time_gpu(i), (time_host(i)*iter_gpu(i))/(time_gpu(i)*iter_host(i))); 
    i = i+1;
  end
  fprintf('\\hline\n');
  fprintf('\\end{tabular}\n');
  fprintf('\\caption{Computational results for deblurring for image dimension $%d\\times %d$ ($\\lambda$=%.3f)\\label{ta:image.%d}}\n', n, n, lbd, n)
  fprintf('\\end{table}\n\n');
%
end

% only shows the last image
if plotresults
    figure;
    subplot(221);
    image(I);
    title('original image');
    subplot(222)
    image(f);
    title('blurred noisy image');
    subplot(223);
    if solve_host
       image(u_host);
    end
    title('recovered image(CPU)');
    subplot(224);
    if solve_GPU
       image(u_gpu);
    end
    title('recovered image(GPU)');

    colormap(gray(256));
end

