/*
 * Elementary math routines for GPU using CUDA.
 *
 * This code provides some BLAS and O(n) operations for
 * our GPU projects.
 *
 * Author: Sangkyun Lee
 * Last Modified: 2008/11/7
 *
 */

/*
 * Notice:
 *
 * This file includes another file, named "transpose_kernel.cu", 
 * in order to provide matrix transpose operation.
 *
 * The file, "transpose_kernel.cu", is from NVIDIA CUDA SDK,
 * which is subject to NVIDIA ownership and copyrights described therein.
 *
 */

#include "cuda.h"

// DO NOT CHANGE: if you make changes, modify the affected code accordingly.
#define BLK_LEN_1D	128		// this is obsolete now.
#define BLK_LEN_2D	16

// Maximum length of the vector than can be reduced.
#define MAX_SUM_SIZE		8*(1<<20)
#define MAX_PAR_SUM_SIZE	(MAX_SUM_SIZE/BLK_LEN_2D/BLK_LEN_2D/8)

#define NILL (FLOAT)

/*
 * Initialization & Finalization
 *
 * A temporary storage in global memory for the routines 
 * that perform reductionh is managed by these functions.
 */
FLOAT *CM_sum;
FLOAT *CM_par_sum;

void cudaMath_init()
{
    CUDA_CALL( cudaMalloc((void**)&CM_sum, sizeof(FLOAT)*1) );
    CUBLAS_CALL( cublasInit() );

    CUDA_CALL( cudaMalloc((void**)&CM_par_sum, sizeof(FLOAT)*MAX_PAR_SUM_SIZE) );
}

void cudaMath_destroy()
{
    CUDA_CALL( cudaFree(CM_sum) );
    CUBLAS_CALL( cublasShutdown() );

    CUDA_CALL( cudaFree(CM_par_sum) );
}

/*
 * Complex and real conversions
 *
 */
__global__ void __toComplex(Complex* y, const FLOAT* x, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;
    Complex c;

    if(X < m && Y < n) {
	lin_idx = __imul(m,Y) + X;
        c.x = x[lin_idx];
        c.y = 0.f;
        y[lin_idx] = c;
    }
}
__global__ void __realPart(FLOAT* y, const Complex* x, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;

    if(X < m && Y < n) {
	int lin_idx = __imul(m,Y) + X;
	Complex c = x[lin_idx];
        y[lin_idx] = c.x;
    }
}
__global__ void __imagPart(FLOAT* y, const Complex* x, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;

    if(X < m && Y < n) {
	int lin_idx = __imul(m,Y) + X;
	Complex c = x[lin_idx];
        y[lin_idx] = c.y;
    }
}
__global__ void __conj(Complex* y, const Complex* x, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;

    if(X < m && Y < n) {
	int lin_idx = __imul(m,Y) + X;
	Complex c = x[lin_idx];
	c.y = -c.y;
        y[lin_idx] = c;
    }
}

/*
 * Elementary Functions
 *
 */

/* PLUS */
__global__ void __plusMatMat(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w1[lin_idx] + w2[lin_idx];
    }
}

__global__ void __update(FLOAT* x, const FLOAT* d, FLOAT steplen, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	x[lin_idx] = x[lin_idx] + steplen*d[lin_idx];
    }
}

__global__ void __plusMatConst(FLOAT* out, const FLOAT* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w[lin_idx] + c;
    }
}

/* MINUS */
__global__ void __minusMatMat(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w1[lin_idx] - w2[lin_idx];
    }
}
__global__ void __minusMatMat(FLOAT* out, const FLOAT* w1, const FLOAT c1, const FLOAT* w2, const FLOAT c2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = c1*w1[lin_idx] - c2*w2[lin_idx];
    }
}

__global__ void __minusMatConst(FLOAT* out, const FLOAT* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w[lin_idx] - c;
    }
}
__global__ void __minusConstMat(FLOAT* out, FLOAT c, const FLOAT* w, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = c - w[lin_idx];
    }
}

/* TIMES */
__global__ void __timesMatMat(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w1[lin_idx] * w2[lin_idx];
    }
}
__global__ void __timesMatMatC(Complex* out, const Complex* w1, const Complex* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	Complex c1 = w1[lin_idx];
	Complex c2 = w2[lin_idx];
	Complex c;
	c.x = __fmul_rn(c1.x,c2.x) - __fmul_rn(c1.y,c2.y);
	c.y = __fmul_rn(c1.y,c2.x) + __fmul_rn(c1.x,c2.y);
	out[lin_idx] = c;
    }
}

__global__ void __timesMatConst(FLOAT* out, const FLOAT* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = w[lin_idx] * c;
    }
}

/* POWER */
__global__ void __powerMatMat(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = powf(w1[lin_idx],w2[lin_idx]);
    }
}
__global__ void __powerMatConst(FLOAT* out, const FLOAT* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = powf(w[lin_idx],c);
    }
}
__global__ void __powerConstMat(FLOAT* out, FLOAT c, const FLOAT* w, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = powf(c,w[lin_idx]);
    }
}

/* RDIVIDE */
__global__ void __rdivideMatMat(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = __fdividef(w1[lin_idx],w2[lin_idx]);
    }
}
__global__ void __rdivideMatConst(FLOAT* out, const FLOAT* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = __fdividef(w[lin_idx],c);
    }
}
__global__ void __rdivideMatConstC(Complex* out, const Complex* w, FLOAT c, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	Complex i, o;
	i = w[lin_idx];
	o.x = __fdividef(i.x, c);
	o.y = __fdividef(i.y, c);
	out[lin_idx] = o;
    }
}
__global__ void __rdivideConstMat(FLOAT* out, FLOAT c, const FLOAT* w, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = __fdividef(c,w[lin_idx]);
    }
}

/* SQRT */
__global__ void __sqrtMat(FLOAT* out, const FLOAT* w, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __imul(m,Y) + X;
	out[lin_idx] = sqrtf(w[lin_idx]);
    }
}

/* Functions based on reduction */
__global__ void __abs_sum(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = fabs(x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] += fabs(x[lin_index+256]);
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] += fabs(x[lin_index+512]);
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] += fabs(x[lin_index+768]);
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] += fabs(x[lin_index+1024]);
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] += fabs(x[lin_index+1280]);
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] += fabs(x[lin_index+1536]);
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] += fabs(x[lin_index+1792]);
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}
__global__ void __sqr_sum(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = __fsqr_rn(x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+256]);
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+512]);
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+768]);
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+1024]);
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+1280]);
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+1536]);
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] += __fsqr_rn(x[lin_index+1792]);
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}
__global__ void __nz_sum(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = ((x[lin_index] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] += ((x[lin_index+256] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] += ((x[lin_index+512] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] += ((x[lin_index+768] != 0)? 1.f:0.f);
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] += ((x[lin_index+1024] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] += ((x[lin_index+1280] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] += ((x[lin_index+1536] != 0)? 1.f:0.f);
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] += ((x[lin_index+1792] != 0)? 1.f:0.f);
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}
__global__ void __abs_min(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = fabs(x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+256]));
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+512]));
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+768]));
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+1024]));
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+1280]));
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+1536]));
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] = fminf(cache[ty][tx], fabs(x[lin_index+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+8]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+4]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+2]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+1]);
        __syncthreads();

        if(tx==0 && ty==0) {
	    float minv = cache[0][0];
	    for(int i=1; i<BLK_LEN_2D;++i)
		minv = fminf(minv, cache[i][0]);
	    par_sum[blockIdx.x] = minv;
	}
    }
}
__global__ void __abs_max(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = fabs(x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+256]));
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+512]));
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+768]));
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+1024]));
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+1280]));
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+1536]));
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], fabs(x[lin_index+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+8]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+4]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+2]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+1]);
        __syncthreads();

        if(tx==0 && ty==0) {
	    float maxv = cache[0][0];
	    for(int i=1; i<BLK_LEN_2D;++i)
		maxv = fmaxf(maxv, cache[i][0]);
	    par_sum[blockIdx.x] = maxv;
	}
    }
}
__global__ void __nill_sum(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = x[lin_index];
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] += x[lin_index+256];
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] += x[lin_index+512];
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] += x[lin_index+768];
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] += x[lin_index+1024];
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] += x[lin_index+1280];
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] += x[lin_index+1536];
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] += x[lin_index+1792];
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}
__global__ void __nill_min(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = (x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+256]));
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+512]));
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+768]));
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+1024]));
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+1280]));
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+1536]));
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] = fminf(cache[ty][tx], (x[lin_index+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+8]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+4]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+2]);
	cache[ty][tx] = fminf(cache[ty][tx], cache[ty][tx+1]);
        __syncthreads();

        if(tx==0 && ty==0) {
	    float minv = cache[0][0];
	    for(int i=1; i<BLK_LEN_2D;++i)
		minv = fminf(minv, cache[i][0]);
	    par_sum[blockIdx.x] = minv;
	}
    }
}
__global__ void __nill_max(FLOAT* par_sum, const FLOAT* x, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = (x[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+256]));
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+512]));
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+768]));
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+1024]));
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+1280]));
    __syncthreads();
    if(lin_index + 1536< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+1536]));
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] = fmaxf(cache[ty][tx], (x[lin_index+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+8]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+4]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+2]);
	cache[ty][tx] = fmaxf(cache[ty][tx], cache[ty][tx+1]);
        __syncthreads();

        if(tx==0 && ty==0) {
	    float maxv = cache[0][0];
	    for(int i=1; i<BLK_LEN_2D;++i)
		maxv = fmaxf(maxv, cache[i][0]);
	    par_sum[blockIdx.x] = maxv;
	}
    }
}

#define REDUCE_FUNCTION_1D_1(NAME, REDUCE_FIRST, REDUCE_REP, FINAL_FUNC) \
inline float NAME (const FLOAT* x, int m)\
{\
    FLOAT sum_h;\
\
    int block_size = 8*BLK_LEN_2D*BLK_LEN_2D;\
    int sum_size = m;\
    int reduced_size = idivup(m, block_size);\
\
    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);\
    dim3 dimGrid(reduced_size,1);\
\
    REDUCE_FIRST <<<dimGrid, dimBlock>>> (CM_par_sum, x, sum_size);\
    sum_size = reduced_size;\
    reduced_size = idivup(reduced_size, block_size);\
    dimGrid.x = reduced_size;\
\
    while(sum_size > 1)\
    {\
	REDUCE_REP  <<<dimGrid, dimBlock>>> (CM_par_sum, CM_par_sum, sum_size);\
\
	sum_size = reduced_size;\
	reduced_size = idivup(reduced_size, block_size);\
	dimGrid.x = reduced_size;\
    }\
\
    CUDA_CALL( cudaMemcpy(&sum_h, CM_par_sum, sizeof(FLOAT), cudaMemcpyDeviceToHost) );\
\
    return FINAL_FUNC(sum_h);\
}


/* NORM and related functions */

// INNER PRODUCT
inline FLOAT inner(const FLOAT* x, const FLOAT* y, int m)
{
    FLOAT result;
    result = cublasSdot(m, x, 1, y, 1);
    return result;
}

// NORM_1
inline FLOAT norm_1(const FLOAT* x, int m)
{
    FLOAT result;
    result = cublasSasum(m, x, 1);
    return result;
}

// My inner project routine is slower than CUBLAS counterpart
/*
__global__ void __times_2(FLOAT* par_sum, const FLOAT* x, const FLOAT* y, INT m)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_index = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_index < m)
	cache[ty][tx] = __fmul_rn(x[lin_index], y[lin_index]);
    __syncthreads();
    if(lin_index + 256< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+256], y[lin_index+256]);
    __syncthreads();
    if(lin_index + 512< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+512], y[lin_index+512]);
    __syncthreads();
    if(lin_index + 768< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+768], y[lin_index+768]);
    __syncthreads();

    if(lin_index + 1024< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+1024], y[lin_index+1024]);
    __syncthreads();
    if(lin_index + 1280< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+1280], y[lin_index+1280]);
    __syncthreadsnd();
    if(lin_index + 1536< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+1536], y[lin_index+1536]);
    __syncthreads();
    if(lin_index + 1792< m)
	cache[ty][tx] += __fmul_rn(x[lin_index+1792], y[lin_index+1792]);
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}

// my inner product
inline float inner (const FLOAT* x, const FLOAT* y, int m)
{
    FLOAT sum_h;

    int block_size = 8*BLK_LEN_2D*BLK_LEN_2D;
    int sum_size = m;
    int reduced_size = idivup(m, block_size);

    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(reduced_size,1);

    __times_2 <<<dimGrid, dimBlock>>> (CM_par_sum, x, y, sum_size);
    sum_size = reduced_size;
    reduced_size = idivup(reduced_size, block_size);
    dimGrid.x = reduced_size;

    while(sum_size > 1)
    {
	__nill_sum  <<<dimGrid, dimBlock>>> (CM_par_sum, CM_par_sum, sum_size);

	sum_size = reduced_size;
	reduced_size = idivup(reduced_size, block_size);
	dimGrid.x = reduced_size;
    }

    CUDA_CALL( cudaMemcpy(&sum_h, CM_par_sum, sizeof(FLOAT), cudaMemcpyDeviceToHost) );

    return sum_h;
}
*/


// my norm_1 is slower than using CUBLAS
//REDUCE_FUNCTION_1D_1(norm_1, __abs_sum, __nill_sum, NILL)

// SUM
REDUCE_FUNCTION_1D_1(sum, __nill_sum, __nill_sum, NILL);

// SQ_SUM
REDUCE_FUNCTION_1D_1(sqsum, __sqr_sum, __nill_sum, NILL)

// NORM_2
REDUCE_FUNCTION_1D_1(norm_2, __sqr_sum, __nill_sum, sqrtf);

// NORM_INF
REDUCE_FUNCTION_1D_1(norm_inf, __abs_max, __nill_max, NILL);

// NNZ (count the number of nonzeros in a vector)
REDUCE_FUNCTION_1D_1(my_nnz, __nz_sum, __nill_sum, NILL);

/* MIN & MAX */
REDUCE_FUNCTION_1D_1(min, __nill_min, __nill_min, NILL);
REDUCE_FUNCTION_1D_1(max, __nill_max, __nill_max, NILL);

/* Matrix Transpose */
#include "transpose_kernel.cu"
#define __transpose transpose

/*
 * Functions with spatial locality (using textures)
 *
 */

texture<float, 1, cudaReadModeElementType> tex;
texture<float, 1, cudaReadModeElementType> tex1;
texture<float, 1, cudaReadModeElementType> tex2;

/* 2-D DIVERGENCE */
__global__ void __div2(FLOAT* out, int m, int n);
void divergence2(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n, cudaStream_t st=0)
{
//    int size = sizeof(FLOAT)*m*n;

    cudaChannelFormatDesc cd = cudaCreateChannelDesc<FLOAT>();

    // texture binding/unbinding is now performed outside, for efficiency.
//    CUDA_CALL( cudaBindTexture(0, tex1, w1, cd, size) );
//    CUDA_CALL( cudaBindTexture(0, tex2, w2, cd, size) );

    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    //dim3 dimGrid((int)fmaxf(1.f,1.f*m/BLK_LEN_2D), (int)fmaxf(1.f, 1.f*n/BLK_LEN_2D));
    dim3 dimGrid(idivup(m, BLK_LEN_2D), idivup(n, BLK_LEN_2D));

    __div2<<<dimGrid, dimBlock, 0, st>>>(out, m, n);

//    CUDA_CALL( cudaUnbindTexture(tex1) );
//    CUDA_CALL( cudaUnbindTexture(tex2) );
}

__global__ void __div2(FLOAT* out, int m, int n)
{
    int bx = blockIdx.x;
    int by = blockIdx.y;
    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx = m*Y + X;
    int lin_idx_prevy = m*(Y-1) + X;
    int lin_idx_prevx = lin_idx - 1;

    if(lin_idx == 0)
	out[lin_idx] = tex1Dfetch(tex1,lin_idx) + tex1Dfetch(tex2,lin_idx);
    else if(by == 0 && ty == 0)
	out[lin_idx] = tex1Dfetch(tex1,lin_idx) + (tex1Dfetch(tex2,lin_idx) - tex1Dfetch(tex2,lin_idx_prevx));
    else if(bx == 0 && tx == 0)
	out[lin_idx] = (tex1Dfetch(tex1,lin_idx) - tex1Dfetch(tex1,lin_idx_prevy)) + tex1Dfetch(tex2,lin_idx);
    else
	out[lin_idx] = (tex1Dfetch(tex1,lin_idx) - tex1Dfetch(tex1,lin_idx_prevy))
		     + (tex1Dfetch(tex2,lin_idx) - tex1Dfetch(tex2,lin_idx_prevx));
}

/* 2-D GRADIENT */
__global__ void __grad2(FLOAT* gx, FLOAT* gy, int m, int n);
void gradient2(FLOAT* gx, FLOAT* gy, FLOAT* f, int m, int n, cudaStream_t st=0)
{
//    int memsize = sizeof(FLOAT)*m*n;

    cudaChannelFormatDesc cd = cudaCreateChannelDesc(8*sizeof(FLOAT), 0, 0, 0, cudaChannelFormatKindFloat);

//    CUDA_CALL( cudaBindTexture(0, tex, f, cd, memsize) );

    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(idivup(m,BLK_LEN_2D), idivup(n,BLK_LEN_2D));

    __grad2<<<dimGrid, dimBlock, 0, st>>>(gx, gy, m, n);

//    CUDA_CALL( cudaUnbindTexture(tex) );
}
//__global__ void __grad2(FLOAT* gx, FLOAT* gy, TYPE1_TEXTURE tex, int m, int n)
__global__ void __grad2(FLOAT* gx, FLOAT* gy, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx = m*Y + X;
    int lin_idx_nexty = m*(Y+1) + X;
    int lin_idx_nextx = lin_idx + 1;

    FLOAT cur;
    
    if(X < m-1 || Y < n-1)
	cur = tex1Dfetch(tex,lin_idx);

    if(Y == n-1)
	gx[lin_idx] = 0.f;
    else if(Y < n-1)
	gx[lin_idx] = tex1Dfetch(tex,lin_idx_nexty) - cur;

    if(X == m-1)
	gy[lin_idx] = 0.f;
    else if(X < m-1)
	gy[lin_idx] = tex1Dfetch(tex,lin_idx_nextx) - cur;
}




